// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "Entity.hpp"
#include "scene/Scene.hpp"
#include "assert/assert.hpp"
#include "util/Easing.hpp"
#include "component/world/World.hpp"
#include "component/world/Map.hpp"

using namespace Dawn;

void Entity::onInit() {
  assertNotNull(this->getMap(), "Entity map cannot be null.");

  // Listen for turn time
  listeners.push_back(getScene()->onUnpausedUpdate.listen([this](float_t delta){
    if(this->turnTime <= 0.0f) return;
    this->turnTime -= delta;
    if(this->turnTime <= 0) this->turnTime = 0.0f;
  }));

  // Listen for movement
  listeners.push_back(getScene()->onUnpausedUpdate.listen([this](float_t delta){
    if(this->stepTime <= 0.0f) return;
    this->stepTime -= delta * this->stepSpeed;

    glm::vec3 nextPosition = this->tilePosition.toWorldSpace();
    
    if(this->stepTime <= 0) {
      this->getItem()->setLocalPosition(nextPosition);
      this->stepTime = 0.0f;
      this->eventMove.emit();
      this->eventStepEnd.emit();
      return;
    }

    float_t t = Easing::easeInOutSine(1.0f - this->stepTime);
    glm::vec3 lastPosition = this->lastTilePosition.toWorldSpace();
    glm::vec3 newLocal = glm::mix(lastPosition, nextPosition, t);
    this->getItem()->setLocalPosition(newLocal);
    this->eventMove.emit();
  }));

  // Notify map.
  getMap()->entityNotifyInit(this->getItem());
}

void Entity::onDispose() {
  auto map = this->map.lock();
  if(map) map->entityNotifyDispose(this->getItem());
}

std::shared_ptr<Map> Entity::getMap() {
  auto map = this->map.lock();
  assertNotNull(map, "Map cannot be null?");
  return map;
}

std::shared_ptr<World> Entity::getWorld() {
  return this->getMap()->getWorld();
}

struct EntityStepResult Entity::move(
  const enum EntityDirection direction,
  const float_t stepSpeed
) {
  assertFalse(this->isMoving(), "Entity is already moving.");
  struct EntityStepResult result;

  if(this->direction != direction) {
    this->turn(direction);
    result.type = EntityStepResultType::Turn;
    return result;
  }

  auto map = this->getMap();
  struct EntityTilePosition newPosition = (
    entityDirectionGetRelativeTilePosition(
      this->tilePosition,
      direction
    )
  );

  // Check for entity in way.
  auto entityInWay = map->getEntityAt(newPosition);
  if(entityInWay) {
    result.type = EntityStepResultType::EntityInWay;
    result.entityInWay = entityInWay;
    return result;
  }

  // Get the tile at the destination, check for height if we are on stairs, etc.
  // If the tile is not walkable, return early.
  auto chunk = map->getChunkAtTile(newPosition);
  auto tilePos = newPosition.toTilePosition();
  auto tileAt = chunk->getTileAt(tilePos);
  if(tileAt.isSolid()) {
    result.type = EntityStepResultType::SolidTileInWay;
    result.tileInWayChunk = chunk;
    result.tileInWayPosition = tilePos;
    return result;
  }

  // Move the entity to the new tile.
  this->lastTilePosition = this->tilePosition;
  this->tilePosition = newPosition;
  this->stepTime = 1.0f;
  this->stepSpeed = stepSpeed;

  this->eventStepStart.emit();
  return result;
}

void Entity::turn(const enum EntityDirection direction) {
  this->direction = direction;
  this->turnTime = ENTITY_TURN_TIME;
  this->eventTurn.emit();
}

void Entity::setPosition(struct EntityTilePosition pos) {
  this->stepTime = 0.0f;
  this->tilePosition = pos;
  this->lastTilePosition = pos;
  this->getItem()->setLocalPosition(pos.toWorldSpace());
  this->eventMove.emit();
}

bool_t Entity::isMoving() {
  return this->stepTime > 0.0f || this->turnTime > 0.0f;
}