// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "component/world/World.hpp"
#include "component/entity/EntityID.hpp"
#include "component/entity/EntityTilePosition.hpp"
#include "Chunk.hpp"

namespace Dawn {
  class Map : public SceneComponent {
    private:
      std::unordered_map<enum EntityID, std::weak_ptr<Entity>> entities;

      uint32_t mapWidth, mapHeight, mapDepth;
      uint8_t mapChunkWidth, mapChunkHeight, mapChunkDepth;
      uint32_t mapChunkCount;
      std::unique_ptr<std::shared_ptr<Chunk>[]> chunks;
      std::vector<std::shared_ptr<Chunk>> chunkOrder;
      struct ChunkPosition chunkTopLeft;
  
      /**
       * Part of the ECS, this function is called when an entity is initialized.
       * 
       * @param item Scene Item that has been initialized.
       */
      void entityNotifyInit(std::shared_ptr<SceneItem> item);

      /**
       * Part of the ECS, this function is called when an entity is disposed.
       * 
       * @param item Scene Item that has been disposed.
       */
      void entityNotifyDispose(std::shared_ptr<SceneItem> entity);

    public:
      std::weak_ptr<World> world;

      void onInit() override;
      void onDispose() override;

      /**
       * Sets the size of the map.
       * 
       * @param width The width of the map.
       * @param height The height of the map.
       * @param depth The depth of the map.
       * @param chunkWidth The width of the maps chunks.
       * @param chunkHeight The height of the maps chunks.
       * @param chunkDepth The depth of the maps chunks.
       */
      void setSize(
        const uint32_t width,
        const uint32_t height,
        const uint32_t depth,
        const uint8_t chunkWidth,
        const uint8_t chunkHeight,
        const uint8_t chunkDepth
      );

      /**
       * Gets the world that this map belongs to.
       * 
       * @return The world that this map belongs to.
       */
      std::shared_ptr<World> getWorld();

      /**
       * Gets the entity with the given ID.
       * 
       * @param id The ID of the entity to get.
       * @return The entity with the given ID.
       */
      std::shared_ptr<Entity> getEntity(const EntityID id);

      /**
       * Gets the entity at the given position.
       * 
       * @param pos The position to check for an entity.
       * @return The entity at the given position, or nullptr.
       */
      std::shared_ptr<Entity> getEntityAt(const EntityTilePosition &pos);

      /**
       * Gets the chunk at the given position.
       * 
       * @param pos The position of the chunk to get.
       * @return The chunk at the given position.
       */
      std::shared_ptr<Chunk> getChunkAt(const struct ChunkPosition &pos);

      /**
       * Gets the chunk at the given tile position.
       * 
       * @param pos The tile position of the chunk to get.
       * @return The chunk at the given tile position.
       */
      std::shared_ptr<Chunk> getChunkAtTile(const EntityTilePosition &pos);

      /**
       * Gets the tile at the given position.
       * 
       * @param pos The position of the tile to get.
       * @return The tile at the given position.
       */
      struct Tile getTileAt(const EntityTilePosition &pos);

      friend class Entity;
  };
}